
#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_core.h"
#include "http_protocol.h"
#include "util_md5.h"

#ifdef APACHE13
#include "ap_compat.h"
#else
#include "apr_strings.h"
#endif

#define AUTH_COOKIE_NAME "auth_tkt"
#define BACK_ARG_NAME "back"
#define MD5_DIGEST_SZ 32
#define MD5andTSTAMP (MD5_DIGEST_SZ + 8)
#define SEPARATOR '!'
#define SEPARATOR_HEX "%21"
#define REMOTE_USER_DATA_ENV "REMOTE_USER_DATA"
#define REMOTE_USER_TOKENS_ENV "REMOTE_USER_TOKENS"

#define TKT_AUTH_VERSION "2.0.0b3"

/* ----------------------------------------------------------------------- */
/* Per-directory configuration */
typedef struct  {
  char *directory;
  char *login_url;
  char *timeout_url;
  char *unauth_url;
  char *auth_domain;
  char *auth_cookie_name;
  char *back_cookie_name;
  char *back_arg_name;
  apr_array_header_t *auth_token;
  int ignore_ip;
  int require_ssl;
  unsigned long timeout_sec;
  double timeout_refresh;
} auth_tkt_dir_conf;

/* Per-server configuration */
typedef struct {
  const char *md5_key;
  char *docroot;
} auth_tkt_serv_conf;

typedef struct auth_tkt_struct {
  char *uid;
  char *tokens;
  char *user_data;
  unsigned int timestamp;
} auth_tkt;

typedef struct {
  request_rec *r;
  char *cookie;
  char *cookie_name;
} cookie_res;

/* ----------------------------------------------------------------------- */
/* Initializer */
#ifdef APACHE13
void
auth_tkt_version(server_rec *s, pool *p)
{
  ap_add_version_component("mod_auth_tkt/" TKT_AUTH_VERSION);
}

#else
static int
auth_tkt_version(apr_pool_t *p, 
  apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
  ap_add_version_component(p, "mod_auth_tkt/" TKT_AUTH_VERSION);
  return DECLINED;
}
#endif

/* Create per-dir config structures */
static void * 
create_auth_tkt_config(apr_pool_t *p, char* path)
{
  auth_tkt_dir_conf *conf = apr_palloc(p, sizeof(*conf));

  conf->directory = path;
  conf->login_url = NULL;
  conf->timeout_url = NULL;
  conf->unauth_url = NULL;
  conf->auth_domain = NULL;
  conf->auth_token = apr_array_make(p, 0, sizeof (char *));
  conf->auth_cookie_name = AUTH_COOKIE_NAME;
  conf->back_cookie_name = NULL;
  conf->back_arg_name = BACK_ARG_NAME;
  conf->ignore_ip = -1;
  conf->require_ssl = -1;
  conf->timeout_sec = 7200;
  conf->timeout_refresh = .5;
  return conf;  
}

/* Create per-dir config structures */
static void * 
merge_auth_tkt_config(apr_pool_t *p, void* parent_dirv, void* subdirv)
{
  auth_tkt_dir_conf *parent = (auth_tkt_dir_conf *) parent_dirv;
  auth_tkt_dir_conf *subdir = (auth_tkt_dir_conf *) subdirv;
  auth_tkt_dir_conf *conf = apr_palloc(p, sizeof(*conf));

  conf->directory = (subdir->directory) ? subdir->directory : parent->directory;
  conf->login_url = (subdir->login_url) ? subdir->login_url : parent->login_url;
  conf->timeout_url = (subdir->timeout_url) ? subdir->timeout_url : parent->timeout_url;
  conf->unauth_url = (subdir->unauth_url) ? subdir->unauth_url : parent->unauth_url;
  conf->auth_domain = (subdir->auth_domain) ? subdir->auth_domain : parent->auth_domain;
  conf->auth_token = (subdir->auth_token->nelts > 0) ? subdir->auth_token : parent->auth_token;
  conf->auth_cookie_name = (subdir->auth_cookie_name) ? subdir->auth_cookie_name : parent->auth_cookie_name;
  conf->back_cookie_name = (subdir->back_cookie_name) ? subdir->back_cookie_name : parent->back_cookie_name;
  conf->back_arg_name = (subdir->back_arg_name) ? subdir->back_arg_name : parent->back_arg_name;
  conf->ignore_ip = (subdir->ignore_ip >= 0) ? subdir->ignore_ip : parent->ignore_ip;
  conf->require_ssl = (subdir->require_ssl >= 0) ? subdir->require_ssl : parent->require_ssl;
  conf->timeout_sec = (subdir->timeout_sec >= 0) ? subdir->timeout_sec : parent->timeout_sec;
  conf->timeout_refresh = (subdir->timeout_refresh >= 0) ? subdir->timeout_refresh : parent->timeout_refresh;

  return conf;
}

/* Create per-server config structures */
static void *
create_auth_tkt_serv_config(apr_pool_t *p, server_rec* s)
{
  auth_tkt_serv_conf *conf = apr_palloc(p, sizeof(*conf));
  conf->md5_key = NULL;

  return conf;
} 

/* ----------------------------------------------------------------------- */
/* Command-specific functions */

module AP_MODULE_DECLARE_DATA auth_tkt_module;

static const char *
setup_md5key (cmd_parms *cmd, void *cfg, const char *param)
{
  auth_tkt_serv_conf *sconf = ap_get_module_config(cmd->server->module_config, 
    &auth_tkt_module);
  sconf->md5_key = param;
  return NULL;
}

static const char *
set_auth_tkt_token (cmd_parms *cmd, void *cfg, const char *param)
{
  char	**new;
  auth_tkt_dir_conf *conf = (auth_tkt_dir_conf *) cfg;

  new = (char **) apr_array_push(conf->auth_token);
  *new = apr_pstrdup(cmd->pool, param);
  return NULL;
}

static const char *
set_auth_tkt_timeout(cmd_parms *cmd, void *cfg, const char *param)
{
  auth_tkt_dir_conf *conf = (auth_tkt_dir_conf *)cfg;
  
  long int minutes = strtol (param, (char **) NULL, 10);
  
  if (minutes < 0) 
    return ("Timeout must be positive");
  
  if (minutes == LONG_MAX) 
    return ("Integer overflow or invalid number");
  
  conf->timeout_sec = minutes * 60;
  
  return NULL;
}

static const char *
set_auth_tkt_timeout_refresh(cmd_parms *cmd, void *cfg, const char *param)
{
  auth_tkt_dir_conf *conf = (auth_tkt_dir_conf *)cfg;
  
  double refresh = atof(param);

  if (refresh < 0 || refresh > 1) 
    return "Refresh flag must be between 0 and 1";
  
  conf->timeout_refresh = refresh;
  
  return NULL;
}

/* Command table */
static const command_rec auth_tkt_cmds[] =
{
  AP_INIT_TAKE1("TKTAuthLoginURL", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, login_url),
    OR_AUTHCFG, "URL to redirect to if authentication fails"),
  AP_INIT_TAKE1("TKTAuthTimeoutURL", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, timeout_url),
    OR_AUTHCFG, "URL to redirect to if cookie times-out"),
  AP_INIT_TAKE1("TKTAuthUnauthURL", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, unauth_url),
    OR_AUTHCFG, "URL to redirect to if valid user without required token"),
  AP_INIT_TAKE1("TKTAuthCookieName", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, auth_cookie_name),
    OR_AUTHCFG, "name to use for ticket cookie"),
  AP_INIT_TAKE1("TKTAuthDomain", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, auth_domain),
    OR_AUTHCFG, "domain to use in cookies"),
  AP_INIT_TAKE1("TKTAuthBackCookieName", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, back_cookie_name),
    OR_AUTHCFG, "name to use for back cookie (NULL for none)"),
  AP_INIT_TAKE1("TKTAuthBackArgName", ap_set_string_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, back_arg_name),
    OR_AUTHCFG, "name to use for back url argument (NULL for none)"),
  AP_INIT_FLAG("TKTAuthIgnoreIP", ap_set_flag_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, ignore_ip),
    OR_AUTHCFG, "whether to ignore remote IP address in ticket"),
  AP_INIT_FLAG("TKTAuthRequireSSL", ap_set_flag_slot, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, require_ssl),
    OR_AUTHCFG, "whether to require HTTPS contexts for authentication"),
  AP_INIT_ITERATE("TKTAuthToken", set_auth_tkt_token, 
    (void *)APR_OFFSETOF(auth_tkt_dir_conf, auth_token),
    OR_AUTHCFG, "token required to access this area (NULL for none)"),
  AP_INIT_TAKE1("TKTAuthTimeoutMin", set_auth_tkt_timeout, 
   NULL, OR_AUTHCFG, "cookie inactivity timeout, in minutes"),
  AP_INIT_TAKE1("TKTAuthTimeoutRefresh", set_auth_tkt_timeout_refresh, 
   NULL, OR_AUTHCFG, "cookie timeout refresh flag (0-1)"),
  AP_INIT_TAKE1("TKTAuthSecret", setup_md5key, 
   NULL, RSRC_CONF, "secret key to use in MD5 digest"),
  {NULL},
};

/* ----------------------------------------------------------------------- */
/* Support functions */

/* Parse cookie. Returns 1 if valid, and details in *parsed; 0 if not */
static int 
parse_ticket(request_rec *r, char **magic, auth_tkt *parsed)
{
  int sepidx, sep2idx;
  char *ticket = *magic;
  int len = strlen(ticket);
  
  /* For some reason (some clients?), tickets sometimes come in quoted */
  if (ticket[len-1] == '"') ticket[len-1] = 0;
  if (ticket[0] == '"') ticket ++;

  /* Basic length check for min size */
  if (len <= MD5andTSTAMP) return 0; 
  
  /* See if there is a uid/data separator */
  sepidx = ap_ind(ticket, SEPARATOR);
  if (sepidx == -1) {	
    /* Ticket either uri-escaped, base64-escaped, or bogus */
    if (strstr(ticket, SEPARATOR_HEX)) {
      ap_unescape_url(ticket);
      sepidx = ap_ind(ticket, SEPARATOR);
    }
    else {
      /* base64 encoded string always longer than original, so len+1 sufficient */
      char *buf = (char *) apr_palloc(r->pool, len+1);  
      apr_base64_decode(buf, ticket);
      sepidx = ap_ind(buf, SEPARATOR);
      /* If still no sepidx, must be bogus */
      if (sepidx == -1) return 0;
      /* Update ticket and *magic to decoded version */
      ticket = *magic = buf;
    }
    /* Reset len */
    len = strlen(ticket);
  }

  /* Recheck length */
  if (len <= MD5andTSTAMP || sepidx < MD5andTSTAMP) return 0; 

#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
    "TKT parse_ticket decoded ticket: %s", ticket);
#endif
  
  /* Get the user id */
  parsed->uid = apr_palloc(r->pool, sepidx - MD5andTSTAMP + 1);
  memcpy(parsed->uid, &ticket[MD5andTSTAMP], sepidx - MD5andTSTAMP);
  parsed->uid[sepidx - MD5andTSTAMP] = '\0';
  
  /* Check for tokens */
  sep2idx = ap_ind(&ticket[sepidx+1], SEPARATOR);
  if (sep2idx == -1) {
#ifdef DEBUG
    /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
         "TKT parse_ticket: no tokens"); */
#endif
    parsed->tokens = apr_palloc(r->pool, 1);
    *parsed->tokens = '\0';
  }
  else {
    /* Swap sepidx and sep2idx */
    int tmp = sepidx;
    sepidx = tmp + sep2idx + 1;
    sep2idx = tmp;
#ifdef DEBUG
    /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
        "TKT parse_ticket: tokens found - sep2=%d, sep=%d, len=%d", 
	sep2idx, sepidx, len); */
#endif
    /* Copy tokens to parsed->tokens */
    parsed->tokens = apr_palloc(r->pool, sepidx-sep2idx);
    apr_snprintf(parsed->tokens, sepidx-sep2idx, "%s", &ticket[sep2idx+1]);
#ifdef DEBUG
    /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
        "TKT parse_ticket tokens: %s", parsed->tokens); */
#endif
  }

  /* Copy user data to parsed->user_data */
  parsed->user_data = apr_palloc(r->pool, len-sepidx+1);
  apr_snprintf(parsed->user_data, len-sepidx+1, "%s", &ticket[sepidx+1]);
  
  /* Copy timestamp to parsed->timestamp */
  sscanf(&ticket[MD5_DIGEST_SZ], "%8x", &(parsed->timestamp));
  
  return 1;
}

/* Search cookie headers for our ticket */
static int 
cookie_match(void *result, const char *key, const char *cookie)
{
  cookie_res * cr = (cookie_res *) result;
  
  if (cookie != NULL) {
    char *cookie_name, *value, *cookiebuf, *end;
#ifdef DEBUG
  /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, cr->r, 
       "TKT cookie_match, key %s against <%s> (name=%s)",
       key, cookie, cr->cookie_name); */
#endif

    cookie_name = apr_palloc(cr->r->pool, strlen(cr->cookie_name) + 2);
    strncpy(cookie_name, cr->cookie_name, strlen(cr->cookie_name));
    cookie_name[strlen(cr->cookie_name)] = '=';
    cookie_name[strlen(cr->cookie_name) + 1] = '\0';

    if ((value = strstr(cookie, cookie_name))) {
      /* Cookie includes our cookie_name - copy (first) value into cookiebuf */
      value += strlen(cookie_name);
      cookiebuf = apr_pstrdup(cr->r->pool, value);
      end = strchr(cookiebuf, ';');
      if (end) *end = '\0';      /* Ignore anything after the next ; */
      cr->cookie = cookiebuf;
#ifdef DEBUG
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, cr->r, 
          "TKT cookie_match: found %s", cookiebuf);
#endif
      return(0);
    }
  }
#ifdef DEBUG
  /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, cr->r, 
     "TKT cookie_match: NOT found"); */
#endif
  return (1);
}

/* Look for a url ticket */
static char *
get_url_ticket(request_rec *r)
{
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);
  const char *args = NULL;  /* url arguments string  */
  const char *key, *val;
  char *ticket = NULL;

  /* Use main request args if subrequest */
  request_rec *r_main = r->main == NULL ? r : r->main;
  if (r_main->args != NULL) {
    args = apr_pstrdup(r->pool, r_main->args); 
  }

  if (args != NULL) {
#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
      "TKT: looking for ticket in url: <%s>", args);
#endif

    while (*args && (val = ap_getword(r->pool, &args, '&'))) {
      key = ap_getword(r->pool, &val, '=');

      if (strcmp(key,conf->auth_cookie_name) == 0) {
	char *domain, *cookie;
        char *secure_cookie = conf->require_ssl > 0 ? "; secure" : "";
	/* ap_unescape_url((char *)key); */
        /* ap_unescape_url((char *)val); */
#ifdef DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
          "TKT: found url ticket: <%s>", val);
#endif

	/* Default to current domain if auth domain not set */
        domain = conf->auth_domain;
        if (!domain) {
          domain = strchr(ap_get_server_name(r), '.');
        }

	/* Set cookie now for subsequent auth */
	cookie = domain ?
	  apr_psprintf(r->pool, "%s=%s; path=/; domain=%s%s", 
            conf->auth_cookie_name, val, domain, secure_cookie) :
	  apr_psprintf(r->pool, "%s=%s; path=/%s", 
            conf->auth_cookie_name, val, secure_cookie);
        apr_table_setn(r->err_headers_out, "Set-Cookie", cookie);
#ifdef DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
          "TKT: Setting new cookie from url: <%s>", cookie);
#endif

        /* Found ticket - ignore rest of arguments */
        break;
      }
    }
  }

  return ticket;
}

/* Look for a cookie ticket */
static char * 
get_cookie_ticket(request_rec *r)
{
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);

  /* Walk cookie headers looking for matching ticket */
  cookie_res *cr = apr_palloc(r->pool, sizeof(*cr));
  cr->r = r;
  cr->cookie = NULL;
  cr->cookie_name = conf->auth_cookie_name;
  apr_table_do(cookie_match, (void *) cr, r->headers_in, "Cookie", NULL);

  /* Give up if cookie not found or too short */
  if (! cr->cookie || strlen(cr->cookie) < MD5andTSTAMP) return NULL;

  return cr->cookie;
}

/* Generate a ticket digest string from the given details */
static char * 
ticket_digest(request_rec *r, auth_tkt *parsed, unsigned int timestamp)
{
  auth_tkt_serv_conf *sconf =
    ap_get_module_config(r->server->module_config, &auth_tkt_module);
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);
  const char *md5_key = sconf->md5_key;
  char *uid = parsed->uid;
  char *tokens = parsed->tokens;
  char *user_data = parsed->user_data;

  unsigned char *buf = apr_palloc(r->pool, 8 + strlen(md5_key) + strlen(uid) + 1 + strlen(tokens) + 1 + strlen(user_data) + 1);
  unsigned char *buf2 = apr_palloc(r->pool, MD5_DIGEST_SZ + strlen(md5_key));
  char *remote_ip = conf->ignore_ip > 0 ? "0.0.0.0" : r->connection->remote_ip;
  unsigned long ip = ntohl(inet_addr(remote_ip));
  int len = 0;
  char *digest;

  if (timestamp == 0) timestamp = parsed->timestamp;

  /* Fatals */
  if (buf == NULL) return (NULL);
  if (ip == INADDR_NONE) return (NULL);

  /* First 8 bytes for ip address + timestamp */
  buf[0] = (unsigned char ) ((ip & 0xff000000) >> 24);
  buf[1] = (unsigned char ) ((ip & 0xff0000) >> 16);
  buf[2] = (unsigned char ) ((ip & 0xff00) >> 8);
  buf[3] = (unsigned char ) ((ip & 0xff));  
  buf[4] = (unsigned char ) ((timestamp	& 0xff000000) >> 24);
  buf[5] = (unsigned char ) ((timestamp	& 0xff0000) >> 16);
  buf[6] = (unsigned char ) ((timestamp	& 0xff00) >> 8);
  buf[7] = (unsigned char ) ((timestamp	& 0xff));  
  len = 8;
  
  /* Append remaining components to buf */
  strcpy((char *)&buf[len], md5_key);
  len += strlen(md5_key);
  strcpy((char *)&buf[len], uid);
  len += strlen(uid);
  buf[len++] = 0;
  strcpy((char *)&buf[len], tokens);
  len += strlen(tokens);
  buf[len++] = 0;
  strcpy((char *)&buf[len], user_data);
  len += strlen(user_data);
  buf[len] = 0;

  /* Generate the initial digest */
  digest = ap_md5_binary(r->pool, buf, len);

  /* Copy digest + md5_key into buf2 */
  len = MD5_DIGEST_SZ + strlen(md5_key);
  memcpy(buf2, digest, MD5_DIGEST_SZ);
  memcpy(&buf2[MD5_DIGEST_SZ], md5_key, len - MD5_DIGEST_SZ);

  /* Generate the second digest */
  digest = ap_md5_binary(r->pool, buf2, len);

  /* Should be noop, but just in case ... */
  if (strlen(digest) > MD5_DIGEST_SZ) digest[MD5_DIGEST_SZ] = 0;

  return (digest);
}

/* Check if this is a parseable and valid (MD5) ticket
 * Returns 1 if valid, and the parsed ticket in parsed, 0 if not */
static int
valid_ticket(request_rec *r, const char *source, char *ticket, auth_tkt *parsed)
{
  char *digest;

  /* Attempt to parse cookie */
  if (! parse_ticket(r, &ticket, parsed)) {
#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_WARNING, APR_SUCCESS, r, 
      "TKT valid_ticket: unparseable %s ticket found (%s)", source, ticket);
#endif
    return 0;
  }

#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
    "TKT valid_ticket: (parsed) uid %s, tokens %s, user_data %s, ts %d", 
    parsed->uid, parsed->tokens, parsed->user_data, parsed->timestamp);
#endif

  /* Check MD5 hash */
  digest = ticket_digest(r, parsed, 0);
#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
    "TKT valid_ticket: invalid hash! digest %s, ticket %32.32s", digest, ticket);
#endif
  if (memcmp(ticket, digest, MD5_DIGEST_SZ) != 0) return 0;

  return 1;
}

/* Check for required auth tokens 
 * Returns 1 on success, 0 on failure */
static int
check_tokens(request_rec *r, char *tokens)
{
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);
  char *next_parsed_token;
  const char *t = NULL;
  int match = 0; 

  /* Success if no tokens required */
  if (conf->auth_token->nelts == 0 || 
      strcmp(((char **) conf->auth_token->elts)[0], "NULL") == 0) {
    return 1;
  }
  /* Failure if required and no user tokens found */
  if (tokens == NULL || strlen(tokens) == 0) {
    return 0;
  }

  t = apr_pstrdup(r->pool, tokens); 
  
  while (*t && (next_parsed_token = ap_getword(r->pool, &t, ','))) {
    char ** auth_tokens = (char **) conf->auth_token->elts;
    int i;

    for (i=0; i < conf->auth_token->nelts; i++) {
      int token_len = strlen(auth_tokens[i]);
      if (strncmp(auth_tokens[i], next_parsed_token, token_len) == 0 &&
	  next_parsed_token[token_len] == 0) {
	match = 1;
	break;
      }
    }
    if (match) break;
  }

#ifdef DEBUG
  if (! match) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
      "TKT: no matching tokens! (user tokens %s)", tokens);
  }
#endif

  return match;
}

/* Refresh the auth cookie if timeout refresh is set */
static void
refresh_cookie(request_rec *r, auth_tkt *parsed)
{
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);

  /* The timeout refresh is a double between 0 and 1, signifying what
   * proportion of the timeout should be left before we refresh i.e. 
   * 0 means never refresh (hard timeouts); 1 means always refresh;
   * .33 means only refresh if less than a third of the timeout 
   * period remains. */ 
  unsigned int now = time(NULL);
  int remainder = parsed->timestamp + conf->timeout_sec - now;
  double refresh_sec = conf->timeout_refresh * conf->timeout_sec;
#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
    "TKT: timeout %ld, refresh %.3f, remainder %d, refresh_sec %.3f",
	conf->timeout_sec, conf->timeout_refresh, remainder, refresh_sec);
#endif

  /* If less than our refresh_sec threshold, freshen the cookie */
  if (remainder < refresh_sec) {
    char *ticket, *ticket_base64, *cookie;
    char *digest = ticket_digest(r, parsed, now);
    char *domain = ap_strchr(ap_get_server_name(r), '.');
    char *secure_cookie = conf->require_ssl > 0 ? "; secure" : "";
    if (parsed->tokens) {
      ticket = apr_psprintf(r->pool,
        "%s%08x%s%c%s%c%s", 
          digest, now, parsed->uid, 
	  SEPARATOR, parsed->tokens, 
	  SEPARATOR, parsed->user_data);
    }
    else {
      ticket = apr_psprintf(r->pool,
        "%s%08x%s%c%s", 
          digest, now, parsed->uid, SEPARATOR, parsed->user_data);
    }
    ticket_base64 = ap_pbase64encode(r->pool, ticket);
  
    /* Send the cookie */
    cookie = domain ? 
      apr_psprintf(r->pool, 
	"%s=%s; path=/; domain=%s%s", 
	conf->auth_cookie_name, ticket_base64, domain, secure_cookie) :
      apr_psprintf(r->pool, 
	"%s=%s; path=/%s", 
	conf->auth_cookie_name, ticket_base64, secure_cookie);
    apr_table_setn(r->err_headers_out, "Set-Cookie", cookie);

#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
      "TKT: refreshing cookie: %s=%s, domain=%s, secure=%s",
	conf->auth_cookie_name, ticket_base64, domain, secure_cookie);
#endif
  }
}

/* Check whether the given timestamp has timed out 
 * Returns 1 if okay, 0 if timed out */
static int
check_timeout(request_rec *r, auth_tkt *parsed)
{
  char *domain, *timeout_date, *timeout_cookie;
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);
  unsigned int now = time(NULL);
  unsigned int timestamp = parsed->timestamp;
  char *secure_cookie = conf->require_ssl > 0 ? "; secure" : "";

  /* Return OK if no timeout configured */
  if (conf->timeout_sec == 0) return 1;

  /* Check whether timestamp is still fresh */
  if (timestamp + conf->timeout_sec >= now) {
#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
      "TKT: cookie timeout still good: now %d, timeout: %ld, tstamp: %d", 
      now, conf->timeout_sec, timestamp);
#endif

    /* Check whether to refresh the cookie */
    if (conf->timeout_refresh > 0) refresh_cookie(r, parsed);

    return 1;
  }

#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, 
    "TKT: ticket timed out: now %d, timeout: %ld, tstamp: %d", 
    now, conf->timeout_sec, timestamp);
#endif
    
  /* Delete cookie (set expired) if invalid, in case we want to set from url */
  domain = ap_strchr(ap_get_server_name(r), '.');
  timeout_date = ap_ht_time(r->pool, now - 3600, "%a %d %b %Y %T %Z", 0);
  timeout_cookie = domain ?
    apr_psprintf(r->pool,
      "%s=; path=/; domain=%s; expires=%s%s",
      conf->auth_cookie_name, domain, timeout_date, secure_cookie) :
    apr_psprintf(r->pool,
      "%s=; path=/; expires=%s%s",
      conf->auth_cookie_name, timeout_date, secure_cookie);
  apr_table_setn(r->err_headers_out, "Set-Cookie", timeout_cookie);

  return 0;
}

/* Strip specified query args from a url */
static char * 
query_strip(request_rec *r, const char *strip)
{
  const char *args = NULL;  /* url arguments string  */
  const char *key, *val;
  char *new_args = "";
  char *p;

  /* Use main request args if subrequest */
  request_rec *r_main = r->main == NULL ? r : r->main;
  if (r_main->args == NULL || strip == NULL) 
    return NULL;

  args = apr_pstrdup(r->pool, r_main->args); 

  /* Convert all '&' to ';' */
  while ((p = ap_strchr(args, '&')))
    *p = ';';

  /* Split into individual args */
  while (*args && (val = ap_getword(r->pool, &args, ';'))) {
    key = ap_getword(r->pool, &val, '=');

    /* Add to new_args only if key != strip */
    if (strlen(strip) != strlen(key) || strncmp(key,strip,strlen(strip)) != 0) 
      new_args = apr_psprintf(r->pool, "%s&%s=%s", new_args, key, val);
  }

  if (strlen(new_args) > 0)
    return new_args + 1;

  return NULL;
}

/* Hex conversion, from httpd util.c */
static const char c2x_table[] = "0123456789abcdef";
static APR_INLINE unsigned char *
c2x(unsigned what, unsigned char *where)
{
#if APR_CHARSET_EBCDIC
  what = apr_xlate_conv_byte(ap_hdrs_to_ascii, (unsigned char)what);
#endif /*APR_CHARSET_EBCDIC*/
  *where++ = '%';
  *where++ = c2x_table[what >> 4];
  *where++ = c2x_table[what & 0xf];
  return where;
}

/* Extra escaping - variant of httpd util.c ap_escape_path_segment */
static char * 
escape_extras(apr_pool_t *p, const char *segment)
{
  char *copy = apr_palloc(p, 3 * strlen(segment) + 1);
  const unsigned char *s = (const unsigned char *)segment;
  unsigned char *d = (unsigned char *)copy;
  unsigned c;
  
  while ((c = *s)) {
    if (c == '=' || c == '&' || c == ':') {
      d = c2x(c, d);
    }
    else {
      *d++ = c;
    }
    ++s;
  }
  *d = '\0';
  return copy;
}

/* External redirect to the given url, setting back cookie or arg */
static int 
redirect(request_rec *r, char *location)
{
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);

  /* Back cookies always use the current domain */
  char *domain = ap_strchr(ap_get_server_name(r), '.');
  char *back_cookie_name = conf->back_cookie_name;
  char *back_arg_name = conf->back_arg_name;
  char *hostinfo, *back;
  char *url, *cookie;

  /* Get the scheme we use (http or https) and port */
  const char *scheme = ap_http_method(r);
  int port = ap_get_server_port(r);

  /* Strip any auth_cookie_name arguments from the current args */
  char *query = query_strip(r, conf->auth_cookie_name);
  if (query == NULL) {
    query = "";
  }
  else if (strlen(query) > 0) {
    query = apr_psprintf(r->pool, "?%s", query);
  }

  /* Build back URL */
  hostinfo = port == apr_uri_default_port_for_scheme(scheme) ?
    apr_psprintf(r->pool, "%s", ap_get_server_name(r)) :
    apr_psprintf(r->pool, "%s:%d", ap_get_server_name(r), port);
  back = apr_psprintf(r->pool, "%s://%s%s%s", 
    scheme, hostinfo, r->uri, query);

#ifdef DEBUG
  /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "[mod_auth_tkt] back url: %s", back); */
#endif
  
  /* Escape testing */
  back = ap_escape_path_segment(r->pool, back);
  back = escape_extras(r->pool, back);

  /* Set back cookie if name is not null */
  if (back_cookie_name) {
    cookie = domain ?
      apr_psprintf(r->pool, "%s=%s; path=/; domain=%s", 
        back_cookie_name, back, domain) :
      apr_psprintf(r->pool, "%s=%s; path=/", 
        back_cookie_name, back);

    apr_table_setn(r->err_headers_out, "Set-Cookie", cookie);
    url = location;
  }

  /* If back_cookie_name not set, add a back url argument to url */
  else {
    char sep = ap_strchr(location, '?') ? ';' : '?';
    url = apr_psprintf(r->pool, "%s%c%s=%s", 
      location, sep, back_arg_name, back);
  }

#ifdef DEBUG
  /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "[mod_auth_tkt] redirect: %s", url); */
#endif
  apr_table_setn(r->headers_out, "Location", url);
  
  return HTTP_TEMPORARY_REDIRECT;
}

/* ----------------------------------------------------------------------- */
/* Debug routines */
void 
dump_config(request_rec *r) 
{
#ifdef DEBUG
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);
  auth_tkt_serv_conf *sconf =
    ap_get_module_config(r->server->module_config, &auth_tkt_module);

  /* Dump config settings */
  fprintf(stderr,"[ mod_auth_tkt config ]\n");
  fprintf(stderr,"URI: %s\n", r->uri);
  fprintf(stderr,"Filename: %s\n", r->filename);
  fprintf(stderr,"TKTAuthSecret: %s\n", 	sconf->md5_key);
  fprintf(stderr,"directory: %s\n", 		conf->directory);
  fprintf(stderr,"TKTAuthLoginURL: %s\n", 	conf->login_url);
  fprintf(stderr,"TKTAuthTimeoutURL: %s\n", 	conf->timeout_url);
  fprintf(stderr,"TKTAuthUnauthURL: %s\n", 	conf->unauth_url);
  fprintf(stderr,"TKTAuthCookieName: %s\n", 	conf->auth_cookie_name);
  fprintf(stderr,"TKTAuthDomain: %s\n", 	conf->auth_domain);
  fprintf(stderr,"TKTAuthBackCookieName: %s\n",	conf->back_cookie_name);
  fprintf(stderr,"TKTAuthBackArgName: %s\n",	conf->back_arg_name);
  fprintf(stderr,"TKTAuthIgnoreIP: %d\n",	conf->ignore_ip);
  fprintf(stderr,"TKTAuthRequireSSL: %d\n", 	conf->require_ssl);
  fprintf(stderr,"TKTAuthTimeoutMin: %ld\n", 	conf->timeout_sec);
  fprintf(stderr,"TKTAuthTimeoutRefresh: %f\n",	conf->timeout_refresh);
  if (conf->auth_token->nelts > 0) {
    char ** auth_token = (char **) conf->auth_token->elts;
    int i;
    for (i = 0; i < conf->auth_token->nelts; i++) {
      fprintf(stderr, "TKTAuthToken: %s\n", auth_token[i]);
    }
  }

  /* Test query_strip */
  /*
  char *query = query_strip(r, "foo");
  if (query == NULL)
    query = "";
  fprintf(stderr,"query: %s\n", query);
  */

  fflush(stderr);
#endif
}

/* ----------------------------------------------------------------------- */
/* Main ticket authentication */
static int
auth_tkt_check(request_rec *r)
{
  char *domain, *ticket;
  auth_tkt *parsed = apr_palloc(r->pool, sizeof(*parsed));
  auth_tkt_dir_conf *conf = 
    ap_get_module_config(r->per_dir_config, &auth_tkt_module);
  auth_tkt_serv_conf *sconf =
    ap_get_module_config(r->server->module_config, &auth_tkt_module);
  const char *scheme = ap_http_method(r);

/*
#ifdef DEBUG
  dump_config(r);
#endif
*/

  /* Module not configured unless login_url is set */
  if (! conf->login_url) {
    ap_log_rerror(APLOG_MARK, APLOG_WARNING, APR_SUCCESS, r, "TKT: TKTAuthLoginURL missing");
    return DECLINED;
  }
  /* Module misconfigured unless md5_key set */
  if (! sconf->md5_key) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_SUCCESS, r, "TKT: TKTAuthSecret missing");
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  /* Redirect/login if scheme not "https" and require_ssl is set */
  if (conf->require_ssl > 0 && strcmp(scheme,"https") != 0) {
    ap_log_rerror(APLOG_MARK, APLOG_WARNING, APR_SUCCESS, r, 
      "TKT: redirect/login - unsecured request, TKTAuthRequireSSL is on");
    return redirect(r, conf->login_url);
  }

  /* Setup cookie domain - default to current domain if auth_domain not set */
  domain = conf->auth_domain;
  if (!domain) domain = ap_strchr(ap_get_server_name(r), '.');
  
  /* Check for url ticket - either found (accept) or empty (reset/login) */
  ticket = get_url_ticket(r);
  if (! ticket || ! valid_ticket(r, "url", ticket, parsed)) {
    ticket = get_cookie_ticket(r);
    if (! ticket || ! valid_ticket(r, "cookie", ticket, parsed)) {
      return redirect(r, conf->login_url);
    }
  }

  /* Valid ticket, check tokens - redirect/unauthorised if so */
  if (! check_tokens(r, parsed->tokens)) {
    return redirect(r, conf->unauth_url ? conf->unauth_url : conf->login_url);
  }

  /* Valid ticket, check timeout - redirect/timed-out if so */
  if (! check_timeout(r, parsed)) {
    return redirect(r, conf->timeout_url ? conf->timeout_url : conf->login_url);
  }

  /* Setup apache user, auth_type, and environment variables */
#ifdef APACHE13
  r->connection->user = parsed->uid;
  r->connection->ap_auth_type = "Basic";
#else
  r->user = parsed->uid;
  r->ap_auth_type = "Basic";
#endif
  apr_table_set(r->subprocess_env, REMOTE_USER_DATA_ENV,   parsed->user_data);
  apr_table_set(r->subprocess_env, REMOTE_USER_TOKENS_ENV, parsed->tokens);

  return OK;
}

/* ----------------------------------------------------------------------- */
/* Setup main module data structure */

#ifdef APACHE13
/* Apache 1.3 style */

module MODULE_VAR_EXPORT auth_tkt_module = {
  STANDARD_MODULE_STUFF, 
  auth_tkt_version,             /* initializer */
  create_auth_tkt_config,	/* create per-dir    config structures */
  merge_auth_tkt_config,	/* merge  per-dir    config structures */
  create_auth_tkt_serv_config,	/* create per-server config structures */
  NULL,                         /* merge  per-server config structures */
  auth_tkt_cmds,                /* table of config file commands       */
  NULL,                         /* handlers */
  NULL,                         /* filename translation */
  auth_tkt_check,               /* check user_id */
  NULL,                         /* check auth */
  NULL,                         /* check access */
  NULL,                         /* type_checker */
  NULL,                         /* fixups */
  NULL,                         /* logger */
  NULL,                         /* header parser */
  NULL,                         /* chitkt_init */
  NULL,                         /* chitkt_exit */
  NULL                          /* post read-request */
};

#else
/* Apache 2.0 style */

/* Register hooks */
static void
auth_tkt_register_hooks (apr_pool_t *p)
{
  ap_hook_post_config(auth_tkt_version, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_check_user_id(auth_tkt_check, NULL, NULL, APR_HOOK_FIRST);
}

/* Declare and populate the main module data structure */
module AP_MODULE_DECLARE_DATA auth_tkt_module = {
  STANDARD20_MODULE_STUFF, 
  create_auth_tkt_config,	/* create per-dir    config structures */
  merge_auth_tkt_config,	/* merge  per-dir    config structures */
  create_auth_tkt_serv_config,	/* create per-server config structures */
  NULL,                         /* merge  per-server config structures */
  auth_tkt_cmds,                /* table of config file commands       */
  auth_tkt_register_hooks       /* register hooks                      */
};

#endif

/* 
 * vim:sw=2
 * arch-tag: 70090400-e89e-46e1-9c78-1a9ad0cd47eb
 */
